# Cochar - create a random character for Call of Cthulhu RPG 7th ed.
# Copyright (C) 2023  Adam Walkiewicz

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""**Cochar - main module**"""
import json
import random
from typing import List, Tuple, Union

import randname

import cochar
import cochar.character
import cochar.occup
import cochar.skill
import cochar.utils
import cochar.error
import cochar.interface

cochar.set_logging_level("debug")

SKILLS_INTERFACE = cochar.interface.SkillsJSONInterface(
    cochar.SKILLS_DATABASE, cochar.ERA
)
SKILLS_GENERATOR = cochar.skill.SkillsGenerator(SKILLS_INTERFACE)


def create_character(
    year: int,
    country: str,
    first_name: str = cochar.FIRST_NAME,
    last_name: str = cochar.LAST_NAME,
    age: int = cochar.AGE,
    sex: str = cochar.SEX,
    random_mode: bool = False,
    occupation: str = cochar.OCCUPATION,
    skills: cochar.skill.SkillsDict = {},
    occup_type: str = cochar.OCCUPATION_TYPE,
    era: str = cochar.ERA,
    tags: List[str] = cochar.TAGS,
    skills_generator: cochar.skill.SkillsGenerator = SKILLS_GENERATOR,
) -> cochar.character.Character:
    """Main function for creating Character.
    Use this function instead of instantiating Character class.

    :param year: year of the game
    :type year: int
    :param country: country of character's origin
    :type country: str
    :param first_name: character's first name, defaults to ""
    :type first_name: str, optional
    :param last_name: character's last name, defaults to ""
    :type last_name: str, optional
    :param age: character's age, defaults to False
    :type age: int, optional
    :param sex: character's sex, defaults to False
    :type sex: str, optional
    :param random_mode: choose occupation completely randomly, regardless the character's statistics, defaults to "False"
    :type random_mode: bool, optional
    :param occupation: character's occupation return provided occupation as character's occupation if it exists, defaults to "None"
    :type occupation: str, optional
    :param skills: character's skills, defaults to {}
    :type skills: Skills, optional
    :param occup_type: occupation type, defaults to None
    :type occup_type: str, optional
    :param era: occupation era, defaults to None
    :type era: str, optional
    :param tags: occupation tags, defaults to None
    :type tags: List[str], optional
    :raises ValueError: raise if sex is incorrect
    :return: generated character
    :rtype: Character
    """
    weights = cochar.WEIGHTS

    sex = generate_sex(sex)

    age: int = generate_age(year, sex, age)

    if not first_name:
        first_name = generate_first_name(year, sex, country, weights)

    if not last_name:
        last_name = generate_last_name(year, sex, country, weights)

    (
        strength,
        condition,
        size,
        dexterity,
        appearance,
        education,
        intelligence,
        power,
        luck,
        move_rate,
    ) = generate_base_characteristics(age=age)

    occupation = cochar.occup.generate_occupation(
        education=education,
        power=power,
        dexterity=dexterity,
        appearance=appearance,
        strength=strength,
        random_mode=random_mode,
        occupation=occupation,
        occup_type=occup_type,
        era=era,
        tags=tags,
    )

    sanity_points, magic_points, hit_points = calc_derived_attributes(
        power, size, condition
    )

    damage_bonus, build, dodge = calc_combat_characteristics(strength, size, dexterity)

    occupation_points = cochar.occup.calc_occupation_points(
        occupation, education, power, dexterity, appearance, strength
    )
    hobby_points = cochar.occup.calc_hobby_points(intelligence)

    skills = skills_generator.generate_skills(
        occupation, occupation_points, hobby_points, dexterity, education, skills
    )

    dodge = skills.get("dodge", dodge)

    return cochar.character.Character(
        year=year,
        country=country,
        first_name=first_name,
        last_name=last_name,
        age=age,
        sex=sex,
        occupation=occupation,
        strength=strength,
        condition=condition,
        size=size,
        dexterity=dexterity,
        appearance=appearance,
        education=education,
        intelligence=intelligence,
        power=power,
        luck=luck,
        damage_bonus=damage_bonus,
        build=build,
        move_rate=move_rate,
        skills=skills,
        dodge=dodge,
        sanity_points=sanity_points,
        magic_points=magic_points,
        hit_points=hit_points,
    )


def generate_age(year: int, sex: str, age: int = False) -> int:
    """Generate characters age, based on year and sex.
    Return age if age is provided.

    :param year: year of the game
    :type year: int
    :param sex: character's sex
    :type sex: str
    :param age: character's age, defaults to False
    :type age: int, optional
    :return: character's age
    :rtype: int
    """
    if not isinstance(year, int):
        raise cochar.error.InvalidYearValue(year)

    if age:
        return age

    if year < 1950:
        corrected_year = 1950
    else:
        year_index = cochar.utils.narrowed_bisect(cochar.utils.YEAR_RANGE, year)
        corrected_year = cochar.utils.YEAR_RANGE[year_index]

    file_name = f"pop{corrected_year}"

    with open(cochar.POP_PYRAMID_PATH, "r", encoding="utf-8") as json_file:

        def correct_age_range(age_range: Tuple[int, int], max_age: int):
            for i, elem in enumerate(age_range):
                if elem[1] > max_age:
                    return i
            return len(age_range) + 1

        max_age_index = correct_age_range(cochar.utils.AGE_RANGE, cochar.MAX_AGE)
        age_population = cochar.utils.AGE_RANGE[:max_age_index]

        age_weights = json.load(json_file)[file_name][sex][3 : 3 + max_age_index]
        age_range = random.choices(age_population, weights=age_weights)[0]

        age = random.randint(*age_range)

    return age


def generate_base_characteristics(
    age,
    strength: int = 0,
    condition: int = 0,
    size: int = 0,
    dexterity: int = 0,
    appearance: int = 0,
    education: int = 0,
    intelligence: int = 0,
    power: int = 0,
    luck: int = 0,
    move_rate: int = 0,
) -> tuple:
    """Return base characteristics based on age as a tuple.

    Base characteristics:
    1. strength
    2. condition
    3. size
    4. dexterity
    5. appearance
    6. education
    7. intelligence
    8. power
    9. luck
    10. move rate

    :param age: character's age
    :type age: int
    :param strength: strength, defaults to 0
    :type strength: int, optional
    :param condition: condition, defaults to 0
    :type condition: int, optional
    :param size: size, defaults to 0
    :type size: int, optional
    :param dexterity: dexterity, defaults to 0
    :type dexterity: int, optional
    :param appearance: appearance, defaults to 0
    :type appearance: int, optional
    :param education: education, defaults to 0
    :type education: int, optional
    :param intelligence: intelligence, defaults to 0
    :type intelligence: int, optional
    :param power: power, defaults to 0
    :type power: int, optional
    :param move_rate: move rate, defaults to 0
    :type move_rate: int, optional
    :param luck: luck, defaults to 0
    :type luck: int, optional
    :return: (strength, condition, size, dexterity, appearance, education, intelligence, power, luck, move_rate)
    :rtype: tuple
    """
    if strength == 0:
        strength = random.randint(15, 90)
    if condition == 0:
        condition = random.randint(15, 90)
    if size == 0:
        size = random.randint(40, 90)
    if dexterity == 0:
        dexterity = random.randint(15, 90)
    if appearance == 0:
        appearance = random.randint(15, 90)
    if education == 0:
        education = random.randint(40, 90)
    if intelligence == 0:
        intelligence = random.randint(40, 90)
    if power == 0:
        power = random.randint(15, 90)
    if move_rate == 0:
        move_rate = 0
    if luck == 0:
        luck = random.randint(15, 90)
    if age <= 19:
        luck = max(luck, random.randint(15, 90))

    age_range = cochar.utils.narrowed_bisect(cochar.MODIFIERS["age_range"], age)
    mod_char_points = cochar.MODIFIERS["mod_char_points"][age_range]
    mod_app = cochar.MODIFIERS["mod_app"][age_range]
    mod_move_rate = cochar.MODIFIERS["mod_move_rate"][age_range]
    mod_edu = cochar.MODIFIERS["mod_edu"][age_range]

    appearance = subtract_points_from_characteristic(appearance, mod_app)
    strength, condition, dexterity = subtract_points_from_str_con_dex(
        strength, condition, dexterity, mod_char_points
    )
    education = characteristic_test(education, mod_edu)
    move_rate = calc_move_rate(strength, dexterity, size) - mod_move_rate

    return (
        strength,
        condition,
        size,
        dexterity,
        appearance,
        education,
        intelligence,
        power,
        luck,
        move_rate,
    )


def calc_derived_attributes(
    power: int,
    size: int,
    condition: int,
    sanity_points: int = 0,
    magic_points: int = 0,
    hit_points: int = 0,
) -> Tuple[int, int, int]:
    """Based on power, size and condition,
    return sanity, magic and hit points

    :param power: power points
    :type power: int
    :param size: size points
    :type size: int
    :param condition: condition points
    :type condition: int
    :param sanity_points: sanity points, defaults to 0
    :type sanity_points: int, optional
    :param magic_points: magic points, defaults to 0
    :type magic_points: int, optional
    :param hit_points: hit points, defaults to 0
    :type hit_points: int, optional
    :return: (sanity points, magic points, hit points)
    :rtype: tuple
    """
    if sanity_points == 0:
        sanity_points = calc_sanity_points(power)
    if magic_points == 0:
        magic_points = calc_magic_points(power)
    if hit_points == 0:
        hit_points = calc_hit_points(size, condition)

    return sanity_points, magic_points, hit_points


def calc_sanity_points(power: int) -> int:
    """Return sanity points based on power

    :param power: power value
    :type power: int
    :return: sanity points
    :rtype: int
    """
    return power


def calc_magic_points(power: int) -> int:
    """Return magic points based on power

    :param power: power value
    :type power: int
    :return: magic points
    :rtype: int
    """
    return power // 5


def calc_hit_points(size: int, condition: int) -> int:
    """Return hit points based on size and condition.

    :param size: size value
    :type size: int
    :param condition: condition value
    :type condition: int
    :return: hit points
    :rtype: int
    """
    return (size + condition) // 10


def calc_combat_characteristics(
    strength: int,
    size: int,
    dexterity: int,
    damage_bonus: str = "",
    build: int = 0,
    dodge: int = 0,
) -> Tuple[str, int, int]:
    """Based on strength, size and dexterity,
    return combat characteristics such as:
    dame bonus, build, dodge

    :param strength: strength points
    :type strength: int
    :param size: size points
    :type size: int
    :param dexterity: dexterity points
    :type dexterity: int
    :return: (damage bonus, build, dodge)
    :rtype: Tuple(str, int, int)
    """
    if damage_bonus == "":
        damage_bonus = calc_damage_bonus(strength, size)
    if build == 0:
        build = calc_build(strength, size)
    if dodge == 0:
        dodge = calc_dodge(dexterity)

    return damage_bonus, build, dodge


def calc_damage_bonus(strength: int, size: int) -> str:
    """Return damage bonus, based on sum of strength and size.

    f: X -> Y

    X: {64, 84, 124, 164, 204, 283, 364, 444, 524}
    Y: {"-2", "-1", "0", "+1K4", "+1K6", "+2K6", "+3K6", "+4K6", "+5K6",}

    TODO: Increase bonus damage for +1K6 for every 80 points above 524

    :param strength: character's strength
    :type strength: int
    :param size: character's size
    :type size: int
    :return: character's damage bonus
    :rtype: str
    """
    sum_str_siz = strength + size
    combat_range = cochar.utils.narrowed_bisect(
        cochar.VALUE_MATRIX["combat_range"], sum_str_siz
    )
    damage_bonus = cochar.VALUE_MATRIX["damage_bonus"][combat_range]
    return damage_bonus


def calc_build(strength: int, size: int) -> int:
    """Return build based on sum od strength and size.

    f: X -> Y

    X: {64, 84, 124, 164, 204, 283, 364, 444, 524}
    Y: {-2, -1, 0, 1, 2, 3, 4, 5, 6}

    TODO: Increase bonus damage for +1K6 for every 80 points above 524

    :param strength: character's strength
    :type strength: int
    :param size: character's size
    :type size: int
    :return: character's build
    :rtype: int
    """
    sum_str_siz = strength + size
    combat_range = cochar.utils.narrowed_bisect(
        cochar.VALUE_MATRIX["combat_range"], sum_str_siz
    )
    build = cochar.VALUE_MATRIX["build"][combat_range]
    return build


def calc_dodge(dexterity: int) -> int:
    """Return dodge based on dexterity:

    dodge = dexterity // 2

    :param dexterity: character's dexterity
    :type dexterity: int
    :return: character's dodge
    :rtype: int
    """
    return dexterity // 2


def subtract_points_from_characteristic(
    characteristic_points: int, subtract_points: int
) -> int:
    """Subtract points from characteristic points,
    but if result would be zero or below, than return 1.

    :param characteristic_points: _description_
    :type characteristic_points: int
    :param subtract_points: _description_
    :type subtract_points: int
    :return: _description_
    :rtype: int

    >>> education = 50
    >>> points_to_subtract = 10
    >>> subtract_points_from_characteristic(education, points_to_subtract)
    40
    >>> points_to_subtract = 60
    >>> subtract_points_from_characteristic(education, points_to_subtract)
    1
    """
    return (
        characteristic_points - subtract_points
        if characteristic_points > subtract_points
        else 1
    )


def subtract_points_from_str_con_dex(
    strength: int, condition: int, dexterity: int, subtract_points: int
) -> Tuple[int, int, int]:
    """Subtract certain amount of points from strength, condition and
    dexterity, but prevent each of the characteristics to be
    lower than 1.

    :param strength: character's strength
    :type strength: int
    :param condition: character's condition
    :type condition: int
    :param dexterity: character's dexterity
    :type dexterity: int
    :param subtract_points: amount of points to subtract
    :type subtract_points: int
    :return: (strength, condition, dexterity)
    :rtype: Tuple[int, int, int]
    """
    characteristics = {
        "strength": strength,
        "condition": condition,
        "dexterity": dexterity,
    }
    for _ in range(subtract_points):
        characteristic_to_subtract = [
            characteristic
            for characteristic, value in characteristics.items()
            if value != 1
        ]
        if characteristic_to_subtract:
            characteristic = random.choice(characteristic_to_subtract)
            characteristics[characteristic] -= 1
        else:
            break

    return (
        characteristics["strength"],
        characteristics["condition"],
        characteristics["dexterity"],
    )


def characteristic_test(tested_value: int, repetition: int = 1) -> int:
    """Perform characteristic test.

    Roll number between 1 to 100,
    If that number is higher than tested value, than increase tested value with random number,
    between 1 to 10.
    Repeat `repetition` times.
    If result would be higher than 99, return 99

    .. note:
        for skills use skill_test()

    :param tested_value: tested value
    :type tested_value: int
    :param repetition: how many test to perform
    :type repetition: int
    :return: unchanged or increased tested value, but not higher than 99
    :rtype: int
    """
    for _ in range(repetition):
        test = random.randint(1, 100)
        if tested_value < test:
            tested_value += random.randint(1, 10)
    return tested_value if tested_value <= 99 else 99


def calc_move_rate(strength: int, dexterity: int, size: int) -> int:
    """Return move rate base on relations between
    strength, dexterity and size

    if both dexterity and strength are smaller than size,
    return 7
    if both strength and size are higher or equal than size,
    return 9
    else return 8

    :param strength: character's strength
    :type strength: int
    :param dexterity: characters dexterity
    :type dexterity: int
    :param size: character's size
    :type size: int
    :return: move rate
    :rtype: int
    """
    if dexterity < size and strength < size:
        move_rate = 7
    elif strength >= size and dexterity >= size:
        move_rate = 9
    else:
        move_rate = 8

    return move_rate


# TODO: write unit test
def generate_last_name(year: int, sex: str, country: str, weights: bool) -> str:
    """Return random last name based on the given parameters

    .. note:
        For generating only last name, use randname module.

    :param year: year of the data set with names (if data set not available use a closes available data set)
    :type year: int
    :param sex: name gender, available options ['M', 'F', 'N', None]
    :type sex: str
    :param country: name country
    :type country: str
    :param weights: If true, take under account popularity of names. [default: True]
    :type weights: bool
    :return: last name
    :rtype: str
    """
    sex = _verify_and_return_sex(sex, country, name="last_names")
    return randname.last_name(
        year,
        sex,
        country,
        weights,
        database=cochar.DATABASE,
        show_warnings=cochar.SHOW_WARNINGS,
    )


# TODO: write unit test
def generate_first_name(year: int, sex: str, country: str, weights: bool) -> str:
    """Return random first name based on given parameters.

    .. note:
        For generating only first name, use randname module.

    :param year: year of the data set with names (if data set not available use a closes available data set)
    :type year: int
    :param sex: name gender, available options ['M', 'F', 'N', None]
    :type sex: str
    :param country: name country
    :type country: str
    :param weights: if true, take under account popularity of names. [default: True]
    :type weights: bool
    :return: first name
    :rtype: str
    """
    sex = _verify_and_return_sex(sex, country, name="first_names")
    return randname.first_name(
        year,
        sex,
        country,
        weights,
        database=cochar.DATABASE,
        show_warnings=cochar.SHOW_WARNINGS,
    )


# TODO: write unit test
def _verify_and_return_sex(sex: str, country: str, name: str) -> str:
    """Return valid sex, based on sex, country and name.
    if provided sex is invalid, it will be overridden, and
    function return valid sex.

    :param sex: _description_
    :type sex: str
    :param country: _description_
    :type country: str
    :param name: _description_
    :type name: str
    :return: _description_
    :rtype: str
    """
    available_sex = randname.show_data()[country][name]
    if sex not in available_sex:
        if "N" in available_sex:
            sex = "N"
        else:
            sex = random.choice(available_sex)
    return sex


def generate_sex(sex: Union[str, bool] = None) -> str:
    """Generate character's sex

    :param sex: Character's sex, if provided return that value
    :type sex: Union[str, bool]
    :raises ValueError: If provided sex is not in {"M", "F", None}, raise this error
    :return: Character's sex as "M" or "F"
    :rtype: str
    """
    if sex not in cochar.SEX_OPTIONS:
        raise ValueError(f"incorrect sex value: {sex} -> ['M', 'F', None]")

    return random.choice(("M", "F")) if sex is None else sex.upper()



# Cochar - create a random character for Call of Cthulhu RPG 7th ed.
# Copyright (C) 2023  Adam Walkiewicz

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import os
import json
import logging
from pathlib import Path
from typing import Dict, List, Union, Set

import randname

__title__ = "cochar"
__version__ = "1.0.1"
__author__ = "Adam Walkiewicz"
__license__ = "AGPL v3.0"

DEFAULT_LOGGING_LEVEL = logging.ERROR

LOGGING_LEVEL_MAP = {
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warning": logging.WARNING,
    "error": logging.ERROR,
    "critical": logging.CRITICAL,
}

logging_level = DEFAULT_LOGGING_LEVEL

logging.basicConfig(
    format="[%(asctime)s][%(levelname)s][%(filename)s:%(funcName)s:%(lineno)d] %(message)s",
    level=logging_level,
)


def set_logging_level(logging_level=logging_level):
    logging_level = LOGGING_LEVEL_MAP.get(logging_level)
    return logging.basicConfig(
        format="[%(asctime)s][%(levelname)s][%(filename)s:%(funcName)s:%(lineno)d] %(message)s",
        level=logging_level,
    )


SEX_OPTIONS: Set[Union[str, bool]] = {"M", "m", "F", "f", None}
_THIS_FOLDER = os.path.dirname(os.path.abspath(__file__))
POP_PYRAMID_PATH = os.path.abspath(
    os.path.join(_THIS_FOLDER, "data", "popPyramid.json")
)

# modification's table for characteristics:
MODIFIERS = {
    "age_range": [19, 39, 49, 59, 69, 79, 90],
    "mod_char_points": [5, 0, 5, 10, 20, 40, 80],
    "mod_app": [0, 0, 5, 10, 15, 20, 25],
    "mod_move_rate": [0, 0, 1, 2, 3, 4, 5],
    "mod_edu": [0, 1, 2, 3, 4, 4, 4],
}
# Value matrix for setting build
VALUE_MATRIX = {
    "combat_range": [64, 84, 124, 164, 204, 283, 364, 444, 524],
    "build": [-2, -1, 0, 1, 2, 3, 4, 5, 6],
    "damage_bonus": ["-2", "-1", "0", "+1K4", "+1K6", "+2K6", "+3K6", "+4K6", "+5K6"],
}

with open(
    os.path.join(_THIS_FOLDER, "data", "occupations.json"), "r", encoding="utf-8"
) as json_file:
    # Full data of occupations
    OCCUPATIONS_DATA: Dict[str, dict] = dict(json.load(json_file))
    # OCCUPATIONS_DATA.pop("test")
    # Just occupations names in the list
    OCCUPATIONS_LIST: List[str] = list(OCCUPATIONS_DATA.keys())
    # Occupations divided on 5 categories depends on skill point calculation method
    OCCUPATIONS_GROUPS: List[List[str]] = [
        [key for key, value in OCCUPATIONS_DATA.items() if "edu" in value["groups"]],
        [key for key, value in OCCUPATIONS_DATA.items() if "edupow" in value["groups"]],
        [key for key, value in OCCUPATIONS_DATA.items() if "edudex" in value["groups"]],
        [key for key, value in OCCUPATIONS_DATA.items() if "eduapp" in value["groups"]],
        [key for key, value in OCCUPATIONS_DATA.items() if "edustr" in value["groups"]],
    ]

with open(
    os.path.join(_THIS_FOLDER, "data", "settings.json"), "r", encoding="utf-8"
) as json_file:
    settings: dict = json.load(json_file)
    MIN_AGE: int = settings["min_age"]
    MAX_AGE: int = settings["max_age"]
    MAX_SKILL_LEVEL: int = settings["max_skill_level"]
    YEAR: int = settings["year"]
    AGE: int = settings["age"]
    SEX: str = settings["sex"]
    FIRST_NAME: str = settings["first_name"]
    LAST_NAME: str = settings["last_name"]
    COUNTRY: str = settings["country"]
    OCCUPATION: str = settings["occupation"]
    WEIGHTS: bool = settings["weights"]
    SHOW_WARNINGS: bool = settings["show_warnings"]
    OCCUPATION_TYPE: str = settings["occupation_type"]
    ERA: str = settings["era"]
    TAGS: List[str] = settings["tags"]

    DATABASE: str = settings["database"]
    if not DATABASE:
        DATABASE = randname.DATABASE

SKILLS_DATABASE = Path() / _THIS_FOLDER / "data" / "skills.json"

from cochar.cochar import *



# Cochar - create a random character for Call of Cthulhu RPG 7th ed.
# Copyright (C) 2023  Adam Walkiewicz

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from abc import ABC, abstractmethod, abstractproperty
from pathlib import Path
from typing import List, Dict

import os
import json
import itertools


class SkillsDataInterface(ABC):
    def __init__(self, database: Path, era: str):
        self.database = database
        self.era = era

    @abstractmethod
    def get_skills(self) -> Dict[str, int]:
        pass

    @abstractmethod
    def get_skills_from_category(self) -> Dict[str, int]:
        pass

    @abstractmethod
    def get_all_skills_names(self) -> List[str]:
        pass

    @abstractmethod
    def get_categories_names(self) -> List[str]:
        pass

    @abstractmethod
    def get_basic_skills_names(self) -> List[str]:
        pass


class SkillsJSONInterface(SkillsDataInterface):
    def __init__(self, database: Path, era: set = None):
        super().__init__(database, era)
        self.load_data(database)
        self.database = database
        if not era:
            self.era = {"classic-1920", "modern"}
        else:
            self.era = set(era)

    def load_data(self, database) -> None:
        with open(database, "r", encoding="utf-8") as json_file:
            self.skills_data: Dict = json.load(json_file)

    def get_skills(self) -> Dict[str, int]:
        return {
            skill: item["value"]
            for skill, item in self.skills_data.items()
            if set(item["era"]).issuperset(self.era)
        }

    def get_all_skills_names(self) -> List[str]:
        return [
            skill
            for skill, item in self.skills_data.items()
            if set(item["era"]).issuperset(self.era)
        ]

    # TODO: filter out categories that are not in current era
    def get_categories_names(self) -> List[str]:
        return list(
            set(
                itertools.chain(
                    *map(lambda item: item["categories"], self.skills_data.values())
                )
            )
        )

    def get_basic_skills_names(self) -> List:
        skills_basic = [
            skill
            for skill, item in self.skills_data.items()
            if "basic" in self.skills_data[skill]["categories"]
            and set(item["era"]).issuperset(self.era)
        ]
        return skills_basic

    def get_skills_from_category(self, category) -> List[str]:
        return [
            skill
            for skill, item in self.skills_data.items()
            if category in self.skills_data[skill]["categories"]
            and set(item["era"]).issuperset(self.era)
        ]


class SkillsSQLInterface(SkillsDataInterface):
    # TODO: implement SQL interface
    pass


class SkillsRedisInterface(SkillsDataInterface):
    # TODO: implement Redis interface
    pass



# Cochar - create a random character for Call of Cthulhu RPG 7th ed.
# Copyright (C) 2023  Adam Walkiewicz

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""**Occupations**
Occupations is a module that contains functions related
with occupations
"""
import copy
import random
from itertools import compress
from typing import List, Tuple

import cochar
import cochar.skill
import cochar.error


def generate_occupation(
    education: int = 1,
    power: int = 1,
    dexterity: int = 1,
    appearance: int = 1,
    strength: int = 1,
    random_mode: bool = False,
    occupation: str = None,
    occup_type: str = None,
    era: List[str] = None,
    tags: List[str] = None,
) -> str:
    """Return occupation based on:
    education, power, dexterity, appearance and strength.

    :param education: education points, defaults to 1
    :type education: int, optional
    :param power: power points, defaults to 1
    :type power: int, optional
    :param dexterity: dexterity points, defaults to 1
    :type dexterity: int, optional
    :param appearance: appearance points, defaults to 1
    :type appearance: int, optional
    :param strength: strength points, defaults to 1
    :type strength: int, optional
    :param random_mode: ignore edu, pow, dex and str points and return totally random occupation, defaults to False
    :type random_mode: bool, optional
    :param occupation: return specified occupation, defaults to None
    :type occupation: str, optional
    :param occup_type: specify type of occupation to return, defaults to None
    :type occup_type: str, optional
    :param era: specify era of occupation to return, defaults to None
    :type era: str, optional
    :param tags: return occupation with defined tags, defaults to None
    :type tags: List[str], optional
    :raises cochar.error.IncorrectOccupation: when occupation is not in the list of available occupations
    :raises cochar.error.NoneOccupationMeetsCriteria: when searching criteria are not met by any occupation
    :return: occupation name
    :rtype: str
    """
    # TODO: What happen if user provide illegal values, strings or below 0?
    skill_points_groups: List[int] = [
        education * 4,  # 1
        education * 2 + power * 2,  # 2
        education * 2 + dexterity * 2,  # 3
        education * 2 + appearance * 2,  # 4
        education * 2 + strength * 2,  # 5
    ]

    if random_mode:
        return random.choice(cochar.OCCUPATIONS_LIST)

    if occupation:
        if occupation not in cochar.OCCUPATIONS_LIST:
            raise cochar.error.IncorrectOccupation(occupation)
        return occupation

    occupation_groups = copy.deepcopy(cochar.OCCUPATIONS_GROUPS)

    if occup_type:
        for i, group in enumerate(occupation_groups):
            occupation_groups[i] = [
                occup
                for occup in group
                if cochar.OCCUPATIONS_DATA[occup]["type"] == occup_type
            ]

    if era:
        for i, group in enumerate(occupation_groups):
            occupation_groups[i] = [
                occup for occup in group if cochar.OCCUPATIONS_DATA[occup]["era"] in era
            ]

    if tags:
        for i, group in enumerate(occupation_groups):
            occupation_groups[i] = [
                occup
                for occup in group
                if set(tags).issubset(set(cochar.OCCUPATIONS_DATA[occup]["tags"]))
            ]

    filtered_occupation_groups = [
        e for e in compress(occupation_groups, occupation_groups)
    ]
    filtered_skill_points_group = [
        e for e in compress(skill_points_groups, occupation_groups)
    ]

    if not filtered_occupation_groups:
        raise cochar.error.NoneOccupationMeetsCriteria(
            f"None occupation meets following criteria: "
            f"type: {occup_type}, era: {era}, tags: {tags}"
        )

    skill_points: int = max(filtered_skill_points_group)
    candidates_for_occupation: List[str] = random.choice(
        [
            group
            for group, points in zip(
                filtered_occupation_groups, filtered_skill_points_group
            )
            if points == skill_points
        ]
    )
    return random.choice(candidates_for_occupation)


def calc_occupation_points(
    occupation: str,
    education: int,
    power: int,
    dexterity: int,
    appearance: int,
    strength: int,
    occupation_points: int = None,
) -> int:
    """Return occupation points based on occupation, education, power,
    dexterity, appearance and strength.

    If ``occupation_points`` provided, return ``occupation_points``

    :param occupation: occupation points
    :type occupation: str
    :param education: education points
    :type education: int
    :param power: power points
    :type power: int
    :param dexterity: dexterity points
    :type dexterity: int
    :param appearance: appearance points
    :type appearance: int
    :param strength: strength points
    :type strength: int
    :param occupation_points: occupation points, if provided function returns that value instead of calculating it, defaults to None
    :type occupation_points: int, optional
    :return: occupation points for provided occupation
    :rtype: int
    """
    return (
        occupation_points
        if occupation_points is not None
        else cochar.skill.calc_skill_points(
            occupation, education, power, dexterity, appearance, strength
        )
    )


def calc_hobby_points(intelligence: int, hobby_points: int = None) -> int:
    """Return hobby points, based on intelligence.

    occupation points = 2 * intelligence

    If `hobby_points` provided, return `hobby_points`

    :param intelligence: intelligence points
    :type intelligence: int
    :param hobby_points: hobby_points, defaults to None
    :type hobby_points: int, optional
    :return: hobby points
    :rtype: int
    """
    return hobby_points if hobby_points else intelligence * 2


def get_occupation_list():
    return sorted(cochar.OCCUPATIONS_LIST)



# Cochar - create a random character for Call of Cthulhu RPG 7th ed.
# Copyright (C) 2023  Adam Walkiewicz

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
**Utilities for cochar module**

:param TRANSLATION_DICT: dictionary with translations for skills
:param AGE_RANGE: tuple[int, int], contains ranges of possible ages.
:param YEAR_RANGE: tuple[int], available years for last names

**legend**:

- a: art/craft
- s: science
- f: fighting
- g: firearms
- i: interpersonal
- l: language
- *: any

"""
from bisect import bisect_left
from typing import Dict, Tuple, Sequence

TRANSLATION_DICT: Dict[str, str] = {
    "a": "art/craft",
    "s": "science",
    "f": "fighting",
    "g": "firearms",
    "i": "interpersonal",
    "l": "language",
    "v": "survival",
    "p": "special",
    "*": None,
}

AGE_RANGE: Tuple[int, int] = (
    (15, 19),
    (20, 24),
    (25, 29),
    (30, 34),
    (35, 39),
    (40, 44),
    (45, 49),
    (50, 54),
    (55, 59),
    (60, 64),
    (65, 69),
    (70, 74),
    (75, 79),
    (80, 84),
    (85, 89),
    (90, 94),
    (95, 99),
)

YEAR_RANGE: Tuple[int] = (
    1950,
    1955,
    1960,
    1965,
    1970,
    1975,
    1980,
    1985,
    1990,
    1995,
    2000,
    2005,
    2010,
    2015,
    2020,
)


def narrowed_bisect(a: Sequence[int], x: int) -> int:
    """Standard bisect_left from bisect module
    can return number that exceed len(a).
    narrowed_bisect returns number that is <= len(a).

    It is to prevent IndexError, as many other variables
    relay on the index number returned.

    :param a: sequence of numbers
    :type a: Sequence
    :param x: number to insert
    :type x: int
    :return: position of insertion
    :rtype: int
    """
    i = bisect_left(a, x)
    return i if i != len(a) else i - 1


def is_skill_valid(skill_value: int) -> bool:
    """Check if skill value is int type and it is not
    below 0.

    :param skill_value: skill value to test
    :type skill_value: int
    :return: True if value is valid, else False
    :rtype: bool
    """
    if not isinstance(skill_value, int):
        return False
    if skill_value < 0:
        return False

    return True



# Cochar - create a random character for Call of Cthulhu RPG 7th ed.
# Copyright (C) 2023  Adam Walkiewicz

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import argparse

from . import cochar


def pars_arguments():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--year",
        type=int,
        required=False,
        default=cochar.YEAR,
        help="Character's year of born",
    )
    parser.add_argument(
        "--first_name",
        type=str,
        required=False,
        default=cochar.FIRST_NAME,
        help="Character's first name",
    )
    parser.add_argument(
        "--last_name",
        type=str,
        required=False,
        default=cochar.LAST_NAME,
        help="Character's last name",
    )
    parser.add_argument(
        "--age",
        type=int,
        required=False,
        default=cochar.AGE,
        help="Character's age",
    )
    parser.add_argument(
        "--sex",
        type=str,
        required=False,
        default=cochar.SEX,
        dest="sex",
        help="Character's sex",
    )
    parser.add_argument(
        "--country",
        type=str,
        required=False,
        default=cochar.COUNTRY,
        choices=["US", "PL", "ES"],
        help="Character's country",
    )
    parser.add_argument(
        "--occupation",
        type=str,
        required=False,
        default=cochar.OCCUPATION,
        help="Character's occupation",
    )
    parser.add_argument(
        "--occup_type",
        type=str,
        required=False,
        default=cochar.OCCUPATION_TYPE,
        choices=["classic", "expansion", "custom"],
        help="Occupation type",
    )
    parser.add_argument(
        "--era",
        type=str,
        required=False,
        default=cochar.ERA,
        choices=["classic-1920", "modern"],
        help="Occupation era",
    )
    parser.add_argument(
        "--tags",
        type=str,
        required=False,
        default=cochar.TAGS,
        choices=["lovecraftian", "criminal"],
        help="Occupation tags",
    )

    return parser.parse_args()


def main():
    args = pars_arguments()
    if args.tags:
        tags = [args.tags]
    else:
        tags = args.tags
    try:
        print(
            cochar.create_character(
                year=args.year,
                first_name=args.first_name,
                last_name=args.last_name,
                age=args.age,
                sex=args.sex,
                country=args.country,
                occupation=args.occupation,
                occup_type=args.occup_type,
                era=args.era,
                tags=tags,
            )
        )
    except cochar.error.NoneOccupationMeetsCriteria as e:
        print(e)


if __name__ == "__main__":
    main()



# Cochar - create a random character for Call of Cthulhu RPG 7th ed.
# Copyright (C) 2023  Adam Walkiewicz

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""**Errors for cochar module**"""
from typing import Any


class CocharError(Exception):
    pass


class IncorrectOccupation(CocharError):
    """Raise when occupation is not in OCCUPATIONS_LIST"""

    def __init__(self, occupation_name: str):
        self.occupation_name = occupation_name
        self.message = f"'{self.occupation_name.capitalize()}' not in available occupations. Check cochar.OCCUPATIONS_LIST"
        super().__init__(self.message)

    def __str__(self):
        return self.message


class CharacteristicValueNotAnInt(CocharError):
    """Characteristic value has to be an integer number"""

    def __init__(self, skill_name: str, skill_value: Any):
        self.skill_name = skill_name
        self.skill_value = skill_value
        self.message = f"{self.skill_name.capitalize()} points must be an integer: '{self.skill_value}'"
        super().__init__(self.message)

    def __str__(self):
        return self.message


class CharacteristicPointsBelowMinValue(CocharError):
    """Characteristic points cannot be below defined minimal value"""

    def __init__(self, skill_name: str, min_value=0):
        self.skill_name = skill_name
        self.min_value = min_value
        self.message = f"'{self.skill_name}' Characteristic points cannot be below: {self.min_value}"
        super().__init__(self.message)

    def __str__(self):
        return self.message


class SkillValueNotAnInt(CocharError):
    """Skill value has to be an integer number"""

    def __init__(self, skill):
        self.skill = skill
        self.message = f"'{self.skill}' Skill value must be an integer"
        super().__init__(self.message)

    def __str__(self):
        return self.message


class SkillPointsBelowZero(CocharError):
    """Skill points cannot be below zero"""

    def __init__(self, skill):
        self.skill = skill
        self.message = f"'{self.skill}' Skill points cannot be below zero"
        super().__init__(self.message)

    def __str__(self):
        return self.message


class SkillsNotADict(CocharError):
    """Skills must be a dictionary"""

    def __init__(self, skill):
        self.skill = skill
        self.message = f"Invalid skills: '{self.skill}'. Skills must be an dictionary"
        super().__init__(self.message)

    def __str__(self):
        return self.message


class InvalidYearValue(CocharError):
    """Year must be an integer number"""

    def __init__(self, year: int):
        self.year = year
        self.message = f"Invalid year: '{self.year}'. Year must be an integer number"
        super().__init__(self.message)

    def __str__(self):
        return self.message


class InvalidSexValue(CocharError):
    """InvalidSexArgument.

    Raise when selected sex is not it available for chosen country.
    """

    def __init__(self, sex: str, available_sex: list):
        self.sex = sex
        self.available_sex = available_sex
        self.message = f"{self.sex} not in {self.available_sex}"
        super().__init__(self.message)

    def __str__(self):
        return f"{self.sex} -> {self.message}."


class EmptyName(CocharError):
    """Name cannot be an empty string"""

    def __str__(self):
        return "Name cannot be an empty string"


class AgeNotInRange(CocharError):
    """Age must be within specified range, default 15 - 90"""

    def __init__(self, age: int, min_age: int, max_age: int):
        self.age = age
        self.min_age = min_age
        self.max_age = max_age
        self.message = (
            f"Age not in range: {self.age} -> [{self.min_age}, {self.max_age}]"
        )
        super().__init__(self.message)

    def __str__(self):
        return f"{self.message}."


class InvalidAgeValue(CocharError):
    """Age must be an integer number"""

    def __init__(self, age: int):
        self.age = age
        self.message = f"'{self.age}' Age must be an integer number"
        super().__init__(self.message)

    def __str__(self):
        return self.message


class InvalidCountryValue(CocharError):
    """Raise when country is not in available countries"""

    def __init__(self, country: str, available_countries: list):
        self.country = country
        self.available_countries = available_countries
        self.message = f"{self.country} not in [{self.available_countries}]"
        super().__init__(self.message)

    def __str__(self):
        return f"{self.country} -> {self.message}."


class InvalidOccupationValue(CocharError):
    """Raise when occupation is not in occupation list"""

    def __init__(self, occupation: str, available_occupations: list):
        self.occupation = occupation
        self.available_occupations = available_occupations
        self.message = f"{self.occupation} not in [{self.available_occupations}]"
        super().__init__(self.message)

    def __str__(self):
        return self.message


class InvalidDamageBonusValue(CocharError):
    """Raise when occupation is not in occupation list"""

    def __init__(self, damage_bonus: str, correct_damage_bonus: list):
        self.damage_bonus = damage_bonus
        self.correct_damage_bonus = correct_damage_bonus
        self.message = f"{self.damage_bonus} not in [{self.correct_damage_bonus}]"
        super().__init__(self.message)

    def __str__(self):
        return self.message


class InvalidBuildValue(CocharError):
    """Raise when occupation is not in occupation list"""

    def __init__(self, build: str, correct_build: list):
        self.build = build
        self.correct_build = correct_build
        self.message = f"{self.build} not in [{self.correct_build}]"
        super().__init__(self.message)

    def __str__(self):
        return self.message


class InvalidOccupationEra(CocharError):
    """Raise when occupation era is not correct"""

    def __init__(self, era: str, correct_era: list):
        self.era = era
        self.correct_era = correct_era
        self.message = f"{self.era} not in [{self.correct_era}]"
        super().__init__(self.message)

    def __str__(self):
        return self.message


class InvalidOccupationType(CocharError):
    """Raise when occupation type is not correct"""

    def __init__(self, _type: str, correct_type: list):
        self._type = _type
        self.correct_type = correct_type
        self.message = f"{self._type} not in [{self.correct_type}]"
        super().__init__(self.message)

    def __str__(self):
        return self.message


class InvalidOccupationTags(CocharError):
    """Raise when tag is not among available tags"""

    def __init__(self, tags: str, correct_tags: list):
        self.tags = tags
        self.correct_tags = correct_tags
        self.message = f"{self.tags} not in [{self.correct_tags}]"
        super().__init__(self.message)

    def __str__(self):
        return self.message


class NoneOccupationMeetsCriteria(CocharError):
    """Raise when searching criteria are not met by any occupation"""

    pass


class OccupationPointsBelowZero(CocharError):
    """Raise when provided occupation points are below 0"""

    pass



# Cochar - create a random character for Call of Cthulhu RPG 7th ed.
# Copyright (C) 2023  Adam Walkiewicz

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
""""This module contains classes related with Character object itself."""

from abc import ABC, abstractmethod
from typing import List, Union

import randname

import cochar
import cochar.cochar
import cochar.error
import cochar.skill


class Validator(ABC):
    """It's a parent class for all other descriptors.

    Defines validate method, that needs to be implemented by
    all children.

    :param ABC: abstract base class
    :type ABC: ABCMeta
    """

    def __set_name__(self, owner, name):
        self.public_name = name
        self.private_name = "_" + name

    def __get__(self, obj, objtype=None):
        return getattr(obj, self.private_name)

    def __set__(self, obj, value):
        self.validate(value)
        setattr(obj, self.private_name, value)

    @abstractmethod
    def validate(self, value):
        pass


class Characteristic(Validator):
    """Base characteristic for character class"""

    def __init__(self, min_value=None):
        self.min_value = min_value

    def validate(self, value: int):
        """Check if characteristic is a valid number and is not below `min_value`

        :param value: value to validate
        :type value: int
        :raises cochar.error.CharacteristicValueNotAnInt: raise if value is not an integer
        :raises cochar.error.CharacteristicPointsBelowZero: raise if value is below min_value
        """
        if not isinstance(value, int):
            raise cochar.error.CharacteristicValueNotAnInt(self.public_name, value)
        if value < self.min_value:
            raise cochar.error.CharacteristicPointsBelowMinValue(value, self.min_value)


class Name(Validator):
    """Character's name"""

    def __set__(self, obj, value):
        self.validate(value)
        setattr(obj, self.private_name, str(value))

    def validate(self, new_name: str):
        """Check if name is a valid name.

        :param new_name: new character name
        :type new_name: str
        :raises cochar.error.EmptyName: raise if `new_name` is empty
        """
        if new_name == "":
            raise cochar.error.EmptyName()
        self._first_name = str(new_name)


class Year(Validator):
    """Year in game.

    Technically any integer is a valid year. But practically that depends
    on database with names used to generate names.

    Besides name, also character's age depends on the year. Probability of
    generating certain year is related with the population age pyramid at the
    certain year.

    Lowest and highest year in the data defines effective range for year. Years
    below or above are treated the same as in the lowest or highest years in data.
    That means for example if data for names ends on the year 2010, than setting
    higher year will give same results as 2010.
    """

    def validate(self, new_year: int) -> None:
        """Validate if year is an integer.

        :param new_year: new year of the game
        :type new_year: int
        :raises cochar.error.InvalidYearValue: raise when `new_year` is not an integer

        >>> character = create_character()
        >>> character.year = 1800
        """
        if not isinstance(new_year, int):
            raise cochar.error.InvalidYearValue(new_year)


class Sex(Validator):
    """Character's sex.

    Available sex options:
    - M: male
    - F: female
    - None: for non binary

    As there are not any data for non binary names.
    When ``None`` is selected sex will be randomly drawn from M or F

    >>> c = Character(year=1925, country="US", sex="F")
    >>> c.sex
    'F'
    """

    def validate(self, new_sex: Union[str, None]) -> None:
        """
        Validate character's sex.

        :param new_sex: character's new sex
        :type new_sex: str | None
        :raises InvalidSexValue: Incorrect sex value: sex -> ['M', 'F', None']
        """
        if new_sex not in cochar.SEX_OPTIONS:
            raise cochar.error.InvalidSexValue(new_sex, cochar.SEX_OPTIONS)


class Age(Validator):
    """Character's age.

    :param min_age: minimal character's age
    :type min_age: int
    :param max_age: maximal character's age
    :type max_age: int
    """

    def __init__(self, min_age: int, max_age: int) -> None:
        self.min_age = min_age
        self.max_age = max_age

    def validate(self, new_age: int) -> None:
        """
        Validate character's age.

        :param new_age: character's new age
        :type new_age: int
        :raises InvalidAgeValue: raise when age is not an integer
        :raises AgeNotInRange: age must be between min and max age
        """
        if not isinstance(new_age, int):
            raise cochar.error.InvalidAgeValue(new_age)

        if not self.min_age <= new_age <= self.max_age:
            raise cochar.error.AgeNotInRange(new_age, self.min_age, self.max_age)


class Country(Validator):
    """Character's country.

    Country depends on available data. By default database from
    external ``randname`` package is taken.
    Country also defines what dataset will be used for generating character's
    name.

    See ``randname.available_countries()``.
    """

    def validate(self, new_country: str) -> None:
        """
        Validate character's country.

        :param new_country: character's new country
        :type new_country: str
        :raises InvalidCountryValue: "Country not available: {new_country} -> {randname.available_countries()}
        """

        available_countries = randname.available_countries()
        if new_country not in available_countries:
            raise cochar.error.InvalidCountryValue(new_country, available_countries)


class Occupation(Validator):
    """Character's occupation.

    Available occupations are defined in occupation database.
    """

    def __init__(self, available_occupations):
        self.available_occupations = available_occupations

    def validate(self, new_occupation: str) -> None:
        """
        Validate character's occupation.

        :param new_occupation: character's new occupation
        :type new_occupation: str
        :raises InvalidOccupationValue: raise when `new_occupation` is not in `OCCUPATION_LIST`
        """
        if new_occupation not in self.available_occupations:
            raise cochar.error.InvalidOccupationValue(
                new_occupation, self.available_occupations
            )


class DamageBonus(Validator):
    """Character damage bonus.

    ``correct_values = ['-2', '-1', '0', '+1K4', '+1K6', '+2K6', '+3K6', '+4K6', '+5K6']``

    """

    def validate(self, new_damage_bonus: str) -> None:
        """
        Validate character's damage bonus.

        :param new_damage_bonus: character's new damage bonus
        :type new_damage_bonus: str
        :raises InvalidDamageBonusValue: Invalid damage bonus. {new_damage_bonus} not in {correct_values}
        """
        # TODO: Increase range. +1 for each 80 point above STR+SIZ
        correct_values = [
            "-2",
            "-1",
            "0",
            "+1K4",
            "+1K6",
            "+2K6",
            "+3K6",
            "+4K6",
            "+5K6",
        ]
        new_damage_bonus = str(new_damage_bonus).upper()
        if new_damage_bonus not in correct_values:
            raise cochar.error.InvalidDamageBonusValue(new_damage_bonus, correct_values)


class Build(Validator):
    """Character's build.

    ``correct_values = [-2, -1, 0, 1, 2, 3, 4, 5, 6]``

    TODO: increase range. +1 for each 80 point above STR+SIZ.
    """

    def validate(self, new_build: int) -> None:
        """
        Validate character's build.

        :param new_build: character's new build
        :type new_build: int
        :raises InvalidBuildValue: Invalid build. {new_build} not in {correct_values}
        """
        correct_values = [-2, -1, 0, 1, 2, 3, 4, 5, 6]
        if new_build not in correct_values:
            raise cochar.error.InvalidBuildValue(new_build, correct_values)


class Character:
    """Container for character.

    .. warning:
        Although this class can be used as standalone class,
        it is advised to use `cochar.create_character()` function
        to generate character.
    """

    strength = Characteristic(min_value=0)
    condition = Characteristic(min_value=0)
    size = Characteristic(min_value=0)
    dexterity = Characteristic(min_value=0)
    appearance = Characteristic(min_value=0)
    education = Characteristic(min_value=0)
    intelligence = Characteristic(min_value=0)
    power = Characteristic(min_value=0)
    luck = Characteristic(min_value=0)
    power = Characteristic(min_value=0)
    move_rate = Characteristic(min_value=0)
    sanity_points = Characteristic(min_value=0)
    magic_points = Characteristic(min_value=0)
    hit_points = Characteristic(min_value=0)
    dodge = Characteristic(min_value=0)

    first_name = Name()
    last_name = Name()

    year = Year()
    sex = Sex()
    age = Age(min_age=cochar.MIN_AGE, max_age=cochar.MAX_AGE)
    country = Country()
    occupation = Occupation(available_occupations=cochar.OCCUPATIONS_LIST)
    damage_bonus = DamageBonus()
    build = Build()

    def __init__(
        self,
        year: int = 0,
        country: str = "",
        first_name: str = "",
        last_name: str = "",
        age: int = 0,
        sex: str = "",
        occupation: str = "",
        strength: int = 0,
        condition: int = 0,
        size: int = 0,
        dexterity: int = 0,
        appearance: int = 0,
        education: int = 0,
        intelligence: int = 0,
        power: int = 0,
        luck: int = 0,
        move_rate: int = 0,
        damage_bonus: str = 0,
        build: int = 0,
        dodge: int = 0,
        skills: cochar.skill.SkillsDict = {},
        sanity_points: int = 0,
        magic_points: int = 0,
        hit_points: int = 0,
    ) -> None:
        self.year = year
        self.country = country
        self.first_name = first_name
        self.last_name = last_name
        self.age = age
        self.sex = sex
        self.occupation = occupation
        self.strength = strength
        self.condition = condition
        self.size = size
        self.dexterity = dexterity
        self.appearance = appearance
        self.education = education
        self.intelligence = intelligence
        self.power = power
        self.move_rate = move_rate
        self.luck = luck
        self.damage_bonus = damage_bonus
        self.build = build
        self.skills = cochar.skill.SkillsDict(skills)
        self.dodge = dodge
        self.sanity_points = sanity_points
        self.magic_points = magic_points
        self.hit_points = hit_points

    @property
    def skills(self) -> cochar.skill.SkillsDict:
        """Character's skills.

        :return: character's skills
        :rtype: cochar.skill.Skills
        """
        return self._skills

    @skills.setter
    def skills(self, new_skills: Union[dict, cochar.skill.SkillsDict]) -> None:
        if isinstance(new_skills, cochar.skill.SkillsDict):
            self._skills = new_skills
        elif isinstance(new_skills, dict):
            self._skills = cochar.skill.SkillsDict(new_skills)
        else:
            raise cochar.error.SkillsNotADict("Invalid skills. Skills must be a dict")

    def get_json_format(self) -> dict:
        """Return character's full characteristics as a dictionary.

        :return: full characteristics
        :rtype: dict
        """
        result = {str(key)[1:]: value for key, value in vars(self).items()}
        result.update({"skills": self.skills.get_json_format()})
        return result

    def __eq__(self, o: object) -> bool:
        return True if self.__dict__ == o.__dict__ else False

    def __repr__(self) -> str:
        return (
            f"Character(year={self._year}, country='{self._country}', "
            f"first_name='{self._first_name}', last_name='{self._last_name}', "
            f"age={self._age}, sex='{self._sex}', occupation='{self._occupation}', "
            f"strength={self._strength}, condition={self._condition}, size={self._size}, "
            f"dexterity={self._dexterity}, appearance={self._appearance}, education={self._education}, "
            f"intelligence={self._intelligence}, power={self._power}, move_rate={self._move_rate}, "
            f"luck={self._luck}, skills={self._skills}, damage_bonus='{self._damage_bonus}', "
            f"build={self._build}, dodge={self._dodge}, sanity_points={self._sanity_points}, "
            f"magic_points={self._magic_points}, hit_points={self._hit_points})"
        )

    def __str__(self) -> str:
        skills = ""
        max_items_in_row = 3
        current_number_of_items_in_row = 0

        for skill, value in self._skills.items():
            if current_number_of_items_in_row == max_items_in_row:
                skills += "\n"
                current_number_of_items_in_row = 0
            skills += f"| {skill.capitalize()}: {value} |"
            current_number_of_items_in_row += 1

        return (
            f"Character\n"
            f"Name: {self._first_name} {self._last_name}\n"
            f"Sex: {self._sex}, Age: {self._age}, Country: {self._country}\n"
            f"Occupation: {self._occupation.capitalize()}\n"
            f"STR: {self._strength} CON: {self._condition} SIZ: {self._size}\n"
            f"DEX: {self._dexterity} APP: {self._appearance} EDU: {self._education}\n"
            f"INT: {self._intelligence} POW: {self._power} Luck: {self._luck}\n"
            f"Damage bonus: {self._damage_bonus}\n"
            f"Build: {self._build}\n"
            f"Dodge: {self._dodge}\n"
            f"Move rate: {self._move_rate}\n"
            f"Skills:\n"
            f"{skills}"
        )



# Cochar - create a random character for Call of Cthulhu RPG 7th ed.
# Copyright (C) 2023  Adam Walkiewicz

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import os
from setuptools import setup

THIS_FOLDER = os.path.dirname(os.path.abspath(__file__))

with open(f"{THIS_FOLDER}/docs/readme.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="cochar",
    version="1.0.1",
    description="Call of Cthulhu character generator",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ajwalkiewicz/cochar",
    project_urls={
        "Documentation": "https://ajwalkiewicz.github.io/cochar/_build/html/index.html"
    },
    author="Adam Walkiewicz",
    license="AGPL v3",
    classifiers=[
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    packages=["cochar"],
    include_package_data=True,
    install_requires=["rname"],
    entry_points={"console_scripts": ["cochar=cochar.__main__:main"]},
)



# Cochar - create a random character for Call of Cthulhu RPG 7th ed.
# Copyright (C) 2023  Adam Walkiewicz

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""**Skills**
Skills module contains all functions related with skills
and Skills object, which is a container for skills

"""
import random
from collections import UserDict
from typing import Dict, List

import cochar
import cochar.error
import cochar.interface
import cochar.utils

Skill = Dict[str, int]


# TODO: write unit test
class SkillsDict(UserDict):
    """Dictionary like object to store character's skills.
    Override __setitem__ to validate skills data.

    :param UserDict: UserDict from collections
    :type UserDict: abc.ABCMeta
    """

    def get_json_format(self):
        """Return Skills as a dictionary"""
        return {k: v for k, v in self.items()}

    def __setitem__(self, key: str, value: int) -> None:
        """Add validation for skill values

        :param key: skill name
        :type key: str
        :param value: skill value
        :type value: int
        :raises SkillValueNotAnInt: when value is not an integer
        :raises SkillPointsBelowZero: when value is less than 0
        """
        key = str(key)
        if not isinstance(value, int):
            raise cochar.error.SkillValueNotAnInt(
                f"Invalid {key.lower()} points. {key.capitalize()} points must be an integer"
            )
        if value < 0:
            raise cochar.error.SkillPointsBelowZero(
                f"{key.capitalize()} points cannot be less than 0"
            )
        self.data[key] = value


class SkillsGenerator:
    def __init__(self, interface: cochar.interface.SkillsDataInterface):
        self.set_interface(interface)

    def set_interface(self, interface: cochar.interface.SkillsDataInterface) -> None:
        self.interface = interface
        self.skills_data = self.interface.get_skills()
        self.skills_all = self.interface.get_all_skills_names()
        self.skills_basic = self.interface.get_basic_skills_names()
        self.skills_categories = self.interface.get_categories_names()

    def generate_skills(
        self,
        occupation: str,
        occupation_points: int,
        hobby_points: int,
        dexterity: int,
        education: int,
        skills: SkillsDict = None,
    ) -> SkillsDict:
        """Return skills based on:
        occupation, occupation_points, hobby_points, dexterity and education

        If `skills` provided, return `skills`

        Each occupation has related skills, that are chosen randomly. Then
        each skill has randomly assigned skill level - number of points related
        with that skill.

        Dexterity is required for `dodge` skill.
        Education is required for `language(own)` skill.

        :param occupation: occupation
        :type occupation: str
        :param occupation_points: occupation_points
        :type occupation_points: int
        :param hobby_points: hobby_points
        :type hobby_points: int
        :param dexterity: dexterity
        :type dexterity: int
        :param education: education
        :type education: int
        :param skills: skills, defaults to None
        :type skills: Skills, optional
        :return: skills with assigned skill level
        :rtype: Skills
        """
        if skills:
            skills = SkillsDict(skills)
        else:
            skills = SkillsDict()

            self.skills_data["dodge"] = dexterity // 2
            self.skills_data["language (own)"] = education

            # Assigning points to credit rating
            credit_rating_points = generate_credit_rating_points(
                occupation, occupation_points
            )

            occupation_points_to_distribute = occupation_points - credit_rating_points

            if occupation_points_to_distribute < 0:
                occupation_points_to_distribute = 0

            default_occupations_skills: list = cochar.OCCUPATIONS_DATA[occupation][
                "skills"
            ].copy()
            occupation_skills_list = self._get_skills_list(default_occupations_skills)

            hobby_skills_list = self._get_skills_list(self.skills_basic)

            skills = self._assign_skill_points(
                occupation_points, occupation_skills_list, skills
            )
            skills = self._assign_skill_points(hobby_points, hobby_skills_list, skills)
            skills = self._filter_skills(skills)

            skills.setdefault("credit rating", credit_rating_points)

        return skills

    def _get_skills_list(self, input_list: list) -> List[str]:
        """Parse an input list taken from `occupations.json` and
        return list of skills

        :param input_list: list of skills from `occupations.json`
        :type input_list: list
        :return: list of skills
        :rtype: List[str]
        """
        skills_list = []
        skills_list += list(
            filter(
                lambda x: len(x) > 2
                and isinstance(x, str)
                and x not in self.skills_categories,
                input_list,
            )
        )
        skills_list += self._get_choice_skills(input_list)
        skills_list += self._get_category_skills(input_list)
        return skills_list

    def _get_choice_skills(self, skills_list: list) -> List[str]:
        """Parse a choice option from skills in `occupation.json` and
        return list of skills

        Example:
        [1, "occult", "natural world"] -> ["occult"]
        It means, return randomly one skills from the following options

        :param skills_list: list of skills to choose
        :type skills_list: list
        :return: list of skills
        :rtype: List[str]
        """
        result = []
        for item in skills_list:
            if isinstance(item, list):
                population = item[1:]
                k = item[0]
                result.extend(random.sample(population, k=k))
                result.extend(self._get_category_skills(result))
                result = list(
                    filter(
                        lambda x: x not in self.interface.get_categories_names()
                        and len(x) > 2,
                        result,
                    )
                )

        return result

    def _get_category_skills(self, skills_list: list) -> List[str]:
        """Parse a category skills, and return list of skills.

        Example:
        "1l" -> ["language (german)"]
        It means, one random language from language group.
        "2*" -> ["first aid", "listen"]
        It means, two random skills of all available skills.

        :param skills_list: list of category skill options
        :type skills_list: list
        :return: list of skills
        :rtype: List[str]
        """
        result = []
        for item in skills_list:
            if len(item) == 2 and isinstance(item, str):
                k = int(item[0])
                if item[1] == "*":
                    population = self.skills_basic
                else:
                    population = self.interface.get_skills_from_category(
                        cochar.utils.TRANSLATION_DICT.get(item[1])
                    )
                result.extend(random.sample(population, k=k))

            elif item in self.skills_categories:
                population = self.interface.get_skills_from_category(item)
                result.extend([random.choice(population)])

        return result

    def _assign_skill_points(
        self, points: int, skills_list: list, skills: SkillsDict
    ) -> SkillsDict:
        """Allocate randomly points to the skills from skills_list
        and store it in Skills object

        :param points: points to allocate
        :type points: int
        :param skills_list: list of skills
        :type skills_list: list
        :param skills: Skills object
        :type skills: Skills
        :return: None
        :rtype: None
        """
        for skill in skills_list:
            if skill in self.skills_all:
                skills.setdefault(skill, self.skills_data[skill])
            else:
                skills.setdefault(skill, 1)

        while points:
            skill = random.choice(skills_list)
            if points <= cochar.MAX_SKILL_LEVEL - skills[skill]:
                points_allocation = random.randint(0, points)
            elif sum(list(skills.values())) % 90 == 0:
                break
            elif skills[skill] >= cochar.MAX_SKILL_LEVEL:
                continue
            else:
                points_allocation = random.randint(
                    0, cochar.MAX_SKILL_LEVEL - skills[skill]
                )
            skills[skill] += points_allocation
            points -= points_allocation

        return skills

    def _filter_skills(self, skills: Dict) -> SkillsDict:
        """Filter out all skills with basic value form given dict.

        >>> example_dict = {'psychoanalysis': 1, 'language (spanish)': 66}
        >>> SkillsGenerator(skills_interface)._filter_skills(example_dict)
        {'language (spanish)': 66}
        """

        def has_skill_default_value(item) -> bool:
            skill, value = item
            default_value = self.skills_data.get(skill)
            if default_value is not None:
                return default_value != value

            return False

        skills = filter(has_skill_default_value, skills.items())

        return SkillsDict(skills)


def generate_credit_rating_points(occupation: str, occupation_points: int) -> int:
    """For provided occupation, and it occupation points, return
    credit rating points.

    :param occupation: occupation
    :type occupation: str
    :param occupation_points: occupation points
    :type occupation_points: int
    :return: credit rating points
    :rtype: int
    """
    credit_rating_range = cochar.OCCUPATIONS_DATA[occupation]["credit_rating"].copy()
    if occupation_points < min(credit_rating_range):
        credit_rating_range = [0, occupation_points]
    if occupation_points < max(credit_rating_range):
        credit_rating_range[1] = occupation_points
    return random.randint(*credit_rating_range)


def calc_skill_points(
    occupation: str,
    education: int,
    power: int,
    dexterity: int,
    appearance: int,
    strength: int,
) -> int:
    """Return skill points based on:
    occupation, education, power, dexterity, appearance and strength.

    Return maximum points for provided occupation.

    :param occupation: occupation name
    :type occupation: str
    :param education: education points
    :type education: int
    :param power: power points
    :type power: int
    :param dexterity: dexterity points
    :type dexterity: int
    :param appearance: appearance points
    :type appearance: int
    :param strength: strength points
    :type strength: int
    :return: skill points
    :rtype: int
    """
    skill_points_groups: tuple[int] = (
        education * 4,  # 1
        education * 2 + power * 2,  # 2
        education * 2 + dexterity * 2,  # 3
        education * 2 + appearance * 2,  # 4
        education * 2 + strength * 2,  # 5
    )
    group_index = [
        index
        for index, group in enumerate(cochar.OCCUPATIONS_GROUPS)
        if occupation in group
    ]
    points = [skill_points_groups[i] for i in group_index]

    return max(points)


def skill_test(tested_value: int, repetition: int = 1) -> int:
    """Perform skill test.

    Works like improvement test. Roll number between 1 to 100,
    If that number is higher than tested value or higher,
    than 95, then increase tested value with random number,
    between 1 to 10.

    Repeat repetition times.

    .. note:
        for characteristics use `characteristic_test()`

    :param tested_value: tested value
    :type tested_value: int
    :param repetition: how many test to perform
    :type repetition: int
    :return: unchanged, or increased tested value
    :rtype: int
    """
    for _ in range(repetition):
        test = random.randint(1, 100)
        if test > tested_value or test > 95:
            tested_value += random.randint(1, 10)
    return tested_value
